from io import BytesIO
from pathlib import Path
from base64 import b64encode
from dataclasses import InitVar, field, dataclass
from typing_extensions import NotRequired, override
from typing import Any, List, Type, Union, Iterable, Optional, TypedDict

from nonebot.adapters import Message as BaseMessage
from nonebot.adapters import MessageSegment as BaseMessageSegment

from .utils import Element, parse, escape


class RawData(TypedDict):
    data: Union[bytes, BytesIO]
    mime: str


class MessageSegment(BaseMessageSegment["Message"]):
    def __str__(self) -> str:
        def _attr(key: str, value: Any):
            if value is True:
                return key
            if value is False:
                return f"no-{key}"
            if isinstance(value, (int, float)):
                return f"{key}={value}"
            return f'{key}="{escape(str(value))}"'

        attrs = " ".join(_attr(k, v) for k, v in self.data.items())
        attrs = f" {attrs}" if attrs else ""
        return f"<{self.type}{attrs}/>"

    @classmethod
    @override
    def get_message_class(cls) -> Type["Message"]:
        return Message

    @staticmethod
    def text(content: str) -> "Text":
        return Text("text", {"text": content})

    @staticmethod
    def at(
        user_id: str,
        name: Optional[str] = None,
    ) -> "At":
        data = {"id": user_id}
        if name:
            data["name"] = name
        return At("at", data)  # type: ignore

    @staticmethod
    def at_role(
        role: str,
        name: Optional[str] = None,
    ) -> "At":
        data = {"role": role}
        if name:
            data["name"] = name
        return At("at_role", data)  # type: ignore

    @staticmethod
    def at_all(here: bool = False) -> "At":
        return At("at", {"type": "here" if here else "all"})

    @staticmethod
    def sharp(channel_id: str, name: Optional[str] = None) -> "Sharp":
        data = {"id": channel_id}
        if name:
            data["name"] = name
        return Sharp("sharp", data)  # type: ignore

    @staticmethod
    def link(href: str, display: Optional[str] = None) -> "Link":
        if display:
            return Link("link", {"text": href, "display": display})
        return Link("link", {"text": href})

    @staticmethod
    def image(
        url: Optional[str] = None,
        path: Optional[Union[str, Path]] = None,
        raw: Optional[RawData] = None,
        extra: Optional[dict] = None,
        cache: Optional[bool] = None,
        timeout: Optional[str] = None,
    ) -> "Image":
        if url:
            data = {"src": url}
        elif path:
            data = {"src": Path(path).as_uri()}
        elif raw:
            bd = raw["data"] if isinstance(raw["data"], bytes) else raw["data"].getvalue()  # type: ignore
            data = {"src": f"data:{raw['mime']};base64,{b64encode(bd).decode()}"}
        else:
            raise ValueError("image need at least one of url, path and raw")
        if cache is not None:
            data["cache"] = cache  # type: ignore
        if timeout is not None:
            data["timeout"] = timeout
        return Image("img", data, extra=extra)  # type: ignore

    @staticmethod
    def audio(
        url: Optional[str] = None,
        path: Optional[Union[str, Path]] = None,
        raw: Optional[RawData] = None,
        extra: Optional[dict] = None,
        cache: Optional[bool] = None,
        timeout: Optional[str] = None,
    ) -> "Audio":
        if url:
            data = {"src": url}
        elif path:
            data = {"src": Path(path).as_uri()}
        elif raw:
            bd = raw["data"] if isinstance(raw["data"], bytes) else raw["data"].getvalue()  # type: ignore
            data = {"src": f"data:{raw['mime']};base64,{b64encode(bd).decode()}"}
        else:
            raise ValueError("audio need at least one of url, path and raw")
        if cache is not None:
            data["cache"] = cache  # type: ignore
        if timeout is not None:
            data["timeout"] = timeout
        return Audio("audio", data, extra=extra)  # type: ignore

    @staticmethod
    def video(
        url: Optional[str] = None,
        path: Optional[Union[str, Path]] = None,
        raw: Optional[RawData] = None,
        extra: Optional[dict] = None,
        cache: Optional[bool] = None,
        timeout: Optional[str] = None,
    ) -> "Video":
        if url:
            data = {"src": url}
        elif path:
            data = {"src": Path(path).as_uri()}
        elif raw:
            bd = raw["data"] if isinstance(raw["data"], bytes) else raw["data"].getvalue()  # type: ignore
            data = {"src": f"data:{raw['mime']};base64,{b64encode(bd).decode()}"}
        else:
            raise ValueError("video need at least one of url, path and raw")
        if cache is not None:
            data["cache"] = cache  # type: ignore
        if timeout is not None:
            data["timeout"] = timeout
        return Video("video", data, extra=extra)  # type: ignore

    @staticmethod
    def file(
        url: Optional[str] = None,
        path: Optional[Union[str, Path]] = None,
        raw: Optional[RawData] = None,
        extra: Optional[dict] = None,
        cache: Optional[bool] = None,
        timeout: Optional[str] = None,
    ) -> "File":
        if url:
            data = {"src": url}
        elif path:
            data = {"src": Path(path).as_uri()}
        elif raw:
            bd = raw["data"] if isinstance(raw["data"], bytes) else raw["data"].getvalue()  # type: ignore
            data = {"src": f"data:{raw['mime']};base64,{b64encode(bd).decode()}"}
        else:
            raise ValueError("file need at least one of url, path and raw")
        if cache is not None:
            data["cache"] = cache  # type: ignore
        if timeout is not None:
            data["timeout"] = timeout
        return File("file", data, extra=extra)  # type: ignore

    @staticmethod
    def bold(text: str) -> "Bold":
        return Bold("bold", {"text": text})

    @staticmethod
    def italic(text: str) -> "Italic":
        return Italic("italic", {"text": text})

    @staticmethod
    def underline(text: str) -> "Underline":
        return Underline("underline", {"text": text})

    @staticmethod
    def strikethrough(text: str) -> "Strikethrough":
        return Strikethrough("strikethrough", {"text": text})

    @staticmethod
    def spoiler(text: str) -> "Spoiler":
        return Spoiler("spoiler", {"text": text})

    @staticmethod
    def code(text: str) -> "Code":
        return Code("code", {"text": text})

    @staticmethod
    def superscript(text: str) -> "Superscript":
        return Superscript("superscript", {"text": text})

    @staticmethod
    def subscript(text: str) -> "Subscript":
        return Subscript("subscript", {"text": text})

    @staticmethod
    def br() -> "Br":
        return Br("br", {"text": "\n"})

    @staticmethod
    def paragraph(text: str) -> "Paragraph":
        return Paragraph("paragraph", {"text": text})

    @staticmethod
    def message(
        mid: Optional[str] = None,
        forward: Optional[bool] = None,
        content: Optional["Message"] = None,
    ) -> "RenderMessage":
        data = {}
        if mid:
            data["id"] = mid
        if forward is not None:
            data["forward"] = forward
        if content:
            data["content"] = content
        return RenderMessage("message", data)  # type: ignore

    @staticmethod
    def quote(
        mid: str,
        forward: Optional[bool] = None,
        content: Optional["Message"] = None,
    ) -> "RenderMessage":
        data = {"id": mid}
        if forward is not None:
            data["forward"] = forward  # type: ignore
        if content:
            data["content"] = content  # type: ignore
        return RenderMessage("quote", data)  # type: ignore

    @staticmethod
    def author(
        user_id: str,
        nickname: Optional[str] = None,
        avatar: Optional[str] = None,
    ) -> "Author":
        data = {"id": user_id}
        if nickname:
            data["nickname"] = nickname
        if avatar:
            data["avatar"] = avatar
        return Author("author", data)  # type: ignore

    @staticmethod
    def action_button(button_id: str, display: Optional[str] = None, theme: Optional[str] = None):
        data = {"type": "action", "id": button_id}
        if display:
            data["display"] = display
        if theme:
            data["theme"] = theme
        return Button("button", data)  # type: ignore

    @staticmethod
    def link_button(url: str, display: Optional[str] = None, theme: Optional[str] = None):
        data = {"type": "link", "href": url}
        if display:
            data["display"] = display
        if theme:
            data["theme"] = theme
        return Button("button", data)  # type: ignore

    @staticmethod
    def input_button(text: str, display: Optional[str] = None, theme: Optional[str] = None):
        data = {"type": "input", "text": text}
        if display:
            data["display"] = display
        if theme:
            data["theme"] = theme
        return Button("button", data)  # type: ignore

    @override
    def is_text(self) -> bool:
        return False


class TextData(TypedDict):
    text: str


@dataclass
class Text(MessageSegment):
    data: TextData = field(default_factory=dict)  # type: ignore

    @override
    def __str__(self) -> str:
        return escape(self.data["text"])

    @override
    def is_text(self) -> bool:
        return True


class AtData(TypedDict):
    id: NotRequired[str]
    name: NotRequired[str]
    role: NotRequired[str]
    type: NotRequired[str]


@dataclass
class At(MessageSegment):
    data: AtData = field(default_factory=dict)  # type: ignore


class SharpData(TypedDict):
    id: str
    name: NotRequired[str]


@dataclass
class Sharp(MessageSegment):
    data: SharpData = field(default_factory=dict)  # type: ignore


class LinkData(TypedDict):
    text: str
    display: NotRequired[str]


@dataclass
class Link(MessageSegment):
    data: LinkData = field(default_factory=dict)  # type: ignore

    @override
    def __str__(self):
        if "display" in self.data:
            return f'<a href="{escape(self.data["text"])}">{escape(self.data["display"])}</a>'
        return f'<a href="{escape(self.data["text"])}"/>'

    @override
    def is_text(self) -> bool:
        return True


class ImageData(TypedDict):
    src: str
    cache: NotRequired[bool]
    timeout: NotRequired[str]
    width: NotRequired[int]
    height: NotRequired[int]


@dataclass
class Image(MessageSegment):
    data: ImageData = field(default_factory=dict)  # type: ignore
    extra: InitVar[Optional[dict]] = None

    def __post_init__(self, extra: Optional[dict]):
        if extra:
            self.data.update(extra)  # type: ignore


class AudioData(TypedDict):
    src: str
    cache: NotRequired[bool]
    timeout: NotRequired[str]


@dataclass
class Audio(MessageSegment):
    data: AudioData = field(default_factory=dict)  # type: ignore
    extra: InitVar[Optional[dict]] = None

    def __post_init__(self, extra: Optional[dict]):
        if extra:
            self.data.update(extra)  # type: ignore


class VideoData(TypedDict):
    src: str
    cache: NotRequired[bool]
    timeout: NotRequired[str]


@dataclass
class Video(MessageSegment):
    data: VideoData = field(default_factory=dict)  # type: ignore
    extra: InitVar[Optional[dict]] = None

    def __post_init__(self, extra: Optional[dict]):
        if extra:
            self.data.update(extra)  # type: ignore


class FileData(TypedDict):
    src: str
    cache: NotRequired[bool]
    timeout: NotRequired[str]


@dataclass
class File(MessageSegment):
    data: FileData = field(default_factory=dict)  # type: ignore
    extra: InitVar[Optional[dict]] = None

    def __post_init__(self, extra: Optional[dict]):
        if extra:
            self.data.update(extra)  # type: ignore


@dataclass
class Bold(MessageSegment):
    data: TextData = field(default_factory=dict)  # type: ignore

    @override
    def __str__(self):
        return f'<b>{escape(self.data["text"])}</b>'

    @override
    def is_text(self) -> bool:
        return True


@dataclass
class Italic(MessageSegment):
    data: TextData = field(default_factory=dict)  # type: ignore

    @override
    def __str__(self):
        return f'<i>{escape(self.data["text"])}</i>'

    @override
    def is_text(self) -> bool:
        return True


@dataclass
class Underline(MessageSegment):
    data: TextData = field(default_factory=dict)  # type: ignore

    @override
    def __str__(self):
        return f'<u>{escape(self.data["text"])}</u>'

    @override
    def is_text(self) -> bool:
        return True


@dataclass
class Strikethrough(MessageSegment):
    data: TextData = field(default_factory=dict)  # type: ignore

    @override
    def __str__(self):
        return f'<s>{escape(self.data["text"])}</s>'

    @override
    def is_text(self) -> bool:
        return True


@dataclass
class Spoiler(MessageSegment):
    data: TextData = field(default_factory=dict)  # type: ignore

    @override
    def __str__(self):
        return f'<spl>{escape(self.data["text"])}</spl>'

    @override
    def is_text(self) -> bool:
        return True


@dataclass
class Code(MessageSegment):
    data: TextData = field(default_factory=dict)  # type: ignore

    @override
    def __str__(self):
        return f'<code>{escape(self.data["text"])}</code>'

    @override
    def is_text(self) -> bool:
        return True


@dataclass
class Superscript(MessageSegment):
    data: TextData = field(default_factory=dict)  # type: ignore

    @override
    def __str__(self):
        return f'<sup>{escape(self.data["text"])}</sup>'

    @override
    def is_text(self) -> bool:
        return True


@dataclass
class Subscript(MessageSegment):
    data: TextData = field(default_factory=dict)  # type: ignore

    @override
    def __str__(self):
        return f'<sub>{escape(self.data["text"])}</sub>'

    @override
    def is_text(self) -> bool:
        return True


@dataclass
class Br(MessageSegment):
    data: TextData = field(default_factory=dict)  # type: ignore

    @override
    def __str__(self):
        return "<br/>"

    @override
    def is_text(self) -> bool:
        return True


@dataclass
class Paragraph(MessageSegment):
    data: TextData = field(default_factory=dict)  # type: ignore

    @override
    def __str__(self):
        return f'<p>{escape(self.data["text"])}</p>'

    @override
    def is_text(self) -> bool:
        return True


class RenderMessageData(TypedDict):
    id: NotRequired[str]
    forward: NotRequired[bool]
    content: NotRequired["Message"]


@dataclass
class RenderMessage(MessageSegment):
    data: RenderMessageData = field(default_factory=dict)  # type: ignore

    @override
    def __str__(self):
        attr = []
        if "id" in self.data:
            attr.append(f'id="{escape(self.data["id"])}"')
        if self.data.get("forward"):
            attr.append("forward")
        if "content" not in self.data:
            return f'<{self.type} {" ".join(attr)} />'
        else:
            return f'<{self.type} {" ".join(attr)}>{self.data["content"]}</{self.type}>'


class AuthorData(TypedDict):
    id: str
    nickname: NotRequired[str]
    avatar: NotRequired[str]


@dataclass
class Author(MessageSegment):
    data: AuthorData = field(default_factory=dict)  # type: ignore


class ButtonData(TypedDict, total=False):
    type: str
    display: NotRequired[str]
    id: NotRequired[str]
    href: NotRequired[str]
    text: NotRequired[str]
    theme: NotRequired[str]


@dataclass
class Button(MessageSegment):
    data: ButtonData = field(default_factory=dict)  # type: ignore

    @override
    def __str__(self):
        attr = [f'type="{escape(self.data["type"])}"']  # type: ignore
        if self.data["type"] == "action":  # type: ignore
            attr.append(f'id="{escape(self.data["id"])}"')  # type: ignore
        if self.data["type"] == "link":  # type: ignore
            attr.append(f'href="{escape(self.data["href"])}"')  # type: ignore
        if self.data["type"] == "input":  # type: ignore
            attr.append(f'text="{escape(self.data["text"])}"')  # type: ignore
        if "theme" in self.data:
            attr.append(f'theme="{escape(self.data["theme"])}"')
        if "display" in self.data:
            return f'<button {" ".join(attr)}>{escape(self.data["display"])}</button>'
        return f'<button {" ".join(attr)} />'


ELEMENT_TYPE_MAP = {
    "text": (Text, "text"),
    "at": (At, "at"),
    "sharp": (Sharp, "sharp"),
    "img": (Image, "img"),
    "image": (Image, "img"),
    "audio": (Audio, "audio"),
    "video": (Video, "video"),
    "file": (File, "file"),
    "author": (Author, "author"),
}

STYLE_TYPE_MAP = {
    "b": (Bold, "bold"),
    "strong": (Bold, "bold"),
    "i": (Italic, "italic"),
    "em": (Italic, "italic"),
    "u": (Underline, "underline"),
    "ins": (Underline, "underline"),
    "s": (Strikethrough, "strikethrough"),
    "del": (Strikethrough, "strikethrough"),
    "spl": (Spoiler, "spoiler"),
    "code": (Code, "code"),
    "sup": (Superscript, "superscript"),
    "sub": (Subscript, "subscript"),
    "p": (Paragraph, "paragraph"),
}


class Message(BaseMessage[MessageSegment]):
    @classmethod
    @override
    def get_segment_class(cls) -> Type[MessageSegment]:
        return MessageSegment

    @override
    def __add__(self, other: Union[str, MessageSegment, Iterable[MessageSegment]]) -> "Message":
        return super().__add__(MessageSegment.text(other) if isinstance(other, str) else other)

    @override
    def __radd__(self, other: Union[str, MessageSegment, Iterable[MessageSegment]]) -> "Message":
        return super().__radd__(MessageSegment.text(other) if isinstance(other, str) else other)

    @staticmethod
    @override
    def _construct(msg: str) -> Iterable[MessageSegment]:
        yield from Message.from_satori_element(parse(msg))

    @classmethod
    def from_satori_element(cls, elements: List[Element]) -> "Message":
        msg = Message()
        for elem in elements:
            if elem.type in ELEMENT_TYPE_MAP:
                seg_cls, seg_type = ELEMENT_TYPE_MAP[elem.type]
                msg.append(seg_cls(seg_type, elem.attrs.copy()))
            elif elem.type in ("a", "link"):
                if elem.children:
                    msg.append(
                        Link("link", {"text": elem.attrs["href"], "display": elem.children[0].attrs["text"]})
                    )
                else:
                    msg.append(Link("link", {"text": elem.attrs["href"]}))
            elif elem.type == "button":
                if elem.children:
                    msg.append(Button("button", {"display": elem.children[0].attrs["text"], **elem.attrs}))  # type: ignore
                else:
                    msg.append(Button("button", {**elem.attrs}))  # type: ignore
            elif elem.type in STYLE_TYPE_MAP:
                seg_cls, seg_type = STYLE_TYPE_MAP[elem.type]
                msg.append(seg_cls(seg_type, {"text": elem.children[0].attrs["text"]}))
            elif elem.type in ("br", "newline"):
                msg.append(Br("br", {"text": "\n"}))
            elif elem.type in ("message", "quote"):
                data = elem.attrs.copy()
                if elem.children:
                    data["content"] = Message.from_satori_element(elem.children)
                msg.append(RenderMessage(elem.type, data))  # type: ignore
            else:
                msg.append(Text("text", {"text": str(elem)}))
        return msg

    @override
    def extract_plain_text(self) -> str:
        return "".join(seg.data["text"] for seg in self if seg.is_text())



from enum import IntEnum
from datetime import datetime
from typing import Any, Dict, List, Union, Generic, Literal, TypeVar, Optional

from pydantic.generics import GenericModel
from pydantic import Extra, Field, BaseModel, validator, root_validator

from .utils import Element, log, parse


class ChannelType(IntEnum):
    TEXT = 0
    VOICE = 1
    CATEGORY = 2
    DIRECT = 3


class Channel(BaseModel, extra=Extra.allow):
    id: str
    type: ChannelType
    name: Optional[str] = None
    parent_id: Optional[str] = None


class Guild(BaseModel, extra=Extra.allow):
    id: str
    name: Optional[str] = None
    avatar: Optional[str] = None


class User(BaseModel, extra=Extra.allow):
    id: str
    name: Optional[str] = None
    nick: Optional[str] = None
    avatar: Optional[str] = None
    is_bot: Optional[bool] = None


class InnerMember(BaseModel, extra=Extra.allow):
    user: Optional[User] = None
    name: Optional[str] = None
    nick: Optional[str] = None
    avatar: Optional[str] = None
    joined_at: Optional[datetime] = None

    @validator("joined_at", pre=True)
    def parse_joined_at(cls, v):
        if v is None:
            return None
        if isinstance(v, datetime):
            return v
        try:
            timestamp = int(v)
        except ValueError as e:
            raise ValueError(f"invalid timestamp: {v}") from e
        return datetime.fromtimestamp(timestamp / 1000)


class OuterMember(InnerMember):
    user: User
    joined_at: datetime


class Role(BaseModel, extra=Extra.allow):
    id: str
    name: Optional[str] = None


class LoginStatus(IntEnum):
    OFFLINE = 0
    ONLINE = 1
    CONNECT = 2
    DISCONNECT = 3
    RECONNECT = 4


class Login(BaseModel, extra=Extra.allow):
    user: Optional[User] = None
    self_id: Optional[str] = None
    platform: Optional[str] = None
    status: LoginStatus


class OuterLogin(Login):
    user: User
    self_id: str
    platform: str


class ArgvInteraction(BaseModel):
    name: str
    arguments: list
    options: Any


class ButtonInteraction(BaseModel):
    id: str


class Opcode(IntEnum):
    EVENT = 0
    PING = 1
    PONG = 2
    IDENTIFY = 3
    READY = 4


class Payload(BaseModel):
    op: Opcode = Field(...)
    body: Optional[Dict[str, Any]] = Field(None)


class Identify(BaseModel):
    token: Optional[str] = None
    sequence: Optional[int] = None


class Ready(BaseModel):
    logins: List[OuterLogin]


class IdentifyPayload(Payload):
    op: Literal[Opcode.IDENTIFY] = Field(Opcode.IDENTIFY)
    body: Identify


class ReadyPayload(Payload):
    op: Literal[Opcode.READY] = Field(Opcode.READY)
    body: Ready


class PingPayload(Payload):
    op: Literal[Opcode.PING] = Field(Opcode.PING)


class PongPayload(Payload):
    op: Literal[Opcode.PONG] = Field(Opcode.PONG)


class InnerMessage(BaseModel, extra=Extra.allow):
    id: str
    content: List[Element]
    channel: Optional[Channel] = None
    guild: Optional[Guild] = None
    member: Optional[InnerMember] = None
    user: Optional[User] = None
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None

    @root_validator(pre=True)
    def ensure_content(cls, values):
        if "content" in values:
            return values
        log(
            "WARNING",
            "received message without content, " "this may be caused by a bug of Satori Server.",
        )
        return {**values, "content": "Unknown"}

    @validator("content", pre=True)
    def parse_content(cls, v):
        if isinstance(v, list):
            return v
        if v is None:
            return None
        if not isinstance(v, str):
            raise ValueError("content must be str")
        return parse(v)

    @validator("created_at", pre=True)
    def parse_created_at(cls, v):
        if v is None:
            return None
        if isinstance(v, datetime):
            return v
        try:
            timestamp = int(v)
        except ValueError as e:
            raise ValueError(f"invalid timestamp: {v}") from e
        return datetime.fromtimestamp(timestamp / 1000)

    @validator("updated_at", pre=True)
    def parse_updated_at(cls, v):
        if v is None:
            return None
        if isinstance(v, datetime):
            return v
        try:
            timestamp = int(v)
        except ValueError as e:
            raise ValueError(f"invalid timestamp: {v}") from e
        return datetime.fromtimestamp(timestamp / 1000)


class OuterMessage(InnerMessage):
    channel: Channel
    guild: Guild
    user: User


class Event(BaseModel, extra=Extra.allow):
    id: int
    type: str
    platform: str
    self_id: str
    timestamp: datetime
    argv: Optional[ArgvInteraction] = None
    button: Optional[ButtonInteraction] = None
    channel: Optional[Channel] = None
    guild: Optional[Guild] = None
    login: Optional[Login] = None
    member: Optional[InnerMember] = None
    message: Optional[InnerMessage] = None
    operator: Optional[User] = None
    role: Optional[Role] = None
    user: Optional[User] = None

    @validator("timestamp", pre=True)
    def parse_timestamp(cls, v):
        if v is None:
            return None
        if isinstance(v, datetime):
            return v
        try:
            timestamp = int(v)
        except ValueError as e:
            raise ValueError(f"invalid timestamp: {v}") from e
        return datetime.fromtimestamp(timestamp / 1000)


class EventPayload(Payload):
    op: Literal[Opcode.EVENT] = Field(Opcode.EVENT)
    body: Event


PayloadType = Union[
    Union[EventPayload, PingPayload, PongPayload, IdentifyPayload, ReadyPayload],
    Payload,
]


T = TypeVar("T")


class PageResult(GenericModel, Generic[T], extra=Extra.allow):
    data: List[T]
    next: Optional[str] = None



import asyncio
from typing_extensions import override
from typing import Any, Dict, List, Literal, Optional

from pydantic import parse_raw_as
from nonebot.utils import escape_tag
from nonebot.exception import WebSocketClosed
from nonebot.drivers import Driver, Request, WebSocket, HTTPClientMixin, WebSocketClientMixin

from nonebot.adapters import Adapter as BaseAdapter

from .bot import Bot
from .utils import API, log
from .config import Config, ClientInfo
from .exception import ApiNotAvailable
from .models import Event as SatoriEvent
from .event import (
    EVENT_CLASSES,
    Event,
    MessageEvent,
    LoginAddedEvent,
    InteractionEvent,
    LoginRemovedEvent,
    LoginUpdatedEvent,
)
from .models import (
    Payload,
    LoginStatus,
    PayloadType,
    PingPayload,
    PongPayload,
    EventPayload,
    ReadyPayload,
    IdentifyPayload,
)


class Adapter(BaseAdapter):
    bots: Dict[str, Bot]

    @override
    def __init__(self, driver: Driver, **kwargs: Any):
        super().__init__(driver, **kwargs)
        # 读取适配器所需的配置项
        self.satori_config: Config = Config.parse_obj(self.config)
        self.tasks: List[asyncio.Task] = []  # 存储 ws 任务
        self.sequences: Dict[str, int] = {}  # 存储 连接序列号
        self.setup()

    @classmethod
    @override
    def get_name(cls) -> str:
        """适配器名称"""
        return "Satori"

    def setup(self) -> None:
        if not isinstance(self.driver, HTTPClientMixin):
            # 判断用户配置的Driver类型是否符合适配器要求，不符合时应抛出异常
            raise RuntimeError(
                f"Current driver {self.config.driver} "
                f"doesn't support http client requests!"
                f"{self.get_name()} Adapter need a HTTPClient Driver to work."
            )
        if not isinstance(self.driver, WebSocketClientMixin):
            raise RuntimeError(
                f"Current driver {self.config.driver} does not support "
                "websocket client! "
                f"{self.get_name()} Adapter need a WebSocketClient Driver to work."
            )
        # 在 NoneBot 启动和关闭时进行相关操作
        self.driver.on_startup(self.startup)
        self.driver.on_shutdown(self.shutdown)

    async def startup(self) -> None:
        """定义启动时的操作，例如和平台建立连接"""
        for client in self.satori_config.satori_clients:
            self.tasks.append(asyncio.create_task(self.ws(client)))

    async def shutdown(self) -> None:
        for task in self.tasks:
            if not task.done():
                task.cancel()

        await asyncio.gather(
            *(asyncio.wait_for(task, timeout=10) for task in self.tasks),
            return_exceptions=True,
        )

    @staticmethod
    def payload_to_json(payload: Payload) -> str:
        return payload.__config__.json_dumps(payload.dict(), default=payload.__json_encoder__)

    async def receive_payload(self, info: ClientInfo, ws: WebSocket) -> Payload:
        payload = parse_raw_as(PayloadType, await ws.receive())
        if isinstance(payload, EventPayload):
            self.sequences[info.identity] = payload.body.id
        return payload

    async def _authenticate(self, info: ClientInfo, ws: WebSocket) -> Optional[Literal[True]]:
        """鉴权连接"""
        payload = IdentifyPayload.parse_obj(
            {
                "body": {
                    "token": info.token,
                },
            }
        )
        if info.identity in self.sequences:
            payload.body.sequence = self.sequences[info.identity]

        try:
            await ws.send(self.payload_to_json(payload))
        except Exception as e:
            log(
                "ERROR",
                "<r><bg #f8bbd0>Error while sending " + "Identify event</bg #f8bbd0></r>",
                e,
            )
            return

        resp = await self.receive_payload(info, ws)
        if not isinstance(resp, ReadyPayload):
            log(
                "ERROR",
                "Received unexpected payload while authenticating: " f"{escape_tag(repr(resp))}",
            )
            return
        for login in resp.body.logins:
            if not login.self_id:
                continue
            if login.status != LoginStatus.ONLINE:
                continue
            if login.self_id not in self.bots:
                bot = Bot(self, login.self_id, login.platform, info)
                self.bot_connect(bot)
                log(
                    "INFO",
                    f"<y>Bot {escape_tag(bot.self_id)}</y> connected",
                )
            else:
                bot = self.bots[login.self_id]
            bot.on_ready(login.user)
        if not self.bots:
            log("WARNING", "No bots connected!")
            return
        return True

    async def _heartbeat(self, info: ClientInfo, ws: WebSocket):
        """心跳"""
        while True:
            log("TRACE", f"Heartbeat {self.sequences[info.identity]}")
            payload = PingPayload.parse_obj({})
            try:
                await ws.send(self.payload_to_json(payload))
            except Exception as e:
                log("WARNING", "Error while sending heartbeat, Ignored!", e)
            await asyncio.sleep(9)

    async def ws(self, info: ClientInfo) -> None:
        ws_url = info.ws_base / "events"
        req = Request("GET", ws_url, timeout=60.0)
        heartbeat_task: Optional["asyncio.Task"] = None
        while True:
            try:
                async with self.websocket(req) as ws:
                    log(
                        "DEBUG",
                        f"WebSocket Connection to " f"{escape_tag(str(ws_url))} established",
                    )
                    try:
                        if not await self._authenticate(info, ws):
                            await asyncio.sleep(3)
                            continue
                        heartbeat_task = asyncio.create_task(self._heartbeat(info, ws))
                        await self._loop(info, ws)
                    except WebSocketClosed as e:
                        log(
                            "ERROR",
                            "<r><bg #f8bbd0>WebSocket Closed</bg #f8bbd0></r>",
                            e,
                        )
                    except Exception as e:
                        log(
                            "ERROR",
                            "<r><bg #f8bbd0>Error while process data from websocket "
                            f"{escape_tag(str(ws_url))}. "
                            f"Trying to reconnect...</bg #f8bbd0></r>",
                            e,
                        )
                    finally:
                        if heartbeat_task:
                            heartbeat_task.cancel()
                            heartbeat_task = None
                        bots = list(self.bots.values())
                        for bot in bots:
                            self.bot_disconnect(bot)
                        bots.clear()
            except Exception as e:
                log(
                    "ERROR",
                    (
                        "<r><bg #f8bbd0>"
                        "Error while setup websocket to "
                        f"{escape_tag(str(ws_url))}. Trying to reconnect..."
                        "</bg #f8bbd0></r>"
                    ),
                    e,
                )
                await asyncio.sleep(3)  # 重连间隔

    async def _loop(self, info: ClientInfo, ws: WebSocket):
        while True:
            payload = await self.receive_payload(info, ws)
            log(
                "TRACE",
                f"Received payload: {escape_tag(repr(payload))}",
            )
            if isinstance(payload, EventPayload):
                try:
                    event = self.payload_to_event(payload.body)
                except Exception as e:
                    log(
                        "WARNING",
                        f"Failed to parse event {escape_tag(repr(payload))}",
                        e,
                    )
                else:
                    if isinstance(event, LoginAddedEvent):
                        bot = Bot(self, event.self_id, event.platform, info)
                        if event.user:
                            bot.on_ready(event.user)
                        self.bot_connect(bot)
                        log(
                            "INFO",
                            f"<y>Bot {escape_tag(bot.self_id)}</y> connected",
                        )
                    elif isinstance(event, LoginRemovedEvent):
                        self.bot_disconnect(self.bots[event.self_id])
                        log(
                            "INFO",
                            f"<y>Bot {escape_tag(event.self_id)}</y> disconnected",
                        )
                        continue
                    elif isinstance(event, LoginUpdatedEvent) and event.user:
                        self.bots[event.self_id].on_ready(event.user)
                    if not (bot := self.bots.get(event.self_id)):
                        log(
                            "WARNING",
                            f"Received event for unknown bot " f"{escape_tag(event.self_id)}",
                        )
                        continue
                    if isinstance(event, (MessageEvent, InteractionEvent)):
                        event = event.convert()
                    asyncio.create_task(bot.handle_event(event))
            elif isinstance(payload, PongPayload):
                log("TRACE", "Pong")
                continue
            else:
                log(
                    "WARNING",
                    f"Unknown payload from server: {escape_tag(repr(payload))}",
                )

    @staticmethod
    def payload_to_event(payload: SatoriEvent) -> Event:
        EventClass = EVENT_CLASSES.get(payload.type, None)
        if EventClass is None:
            log("WARNING", f"Unknown payload type: {payload.type}")
            return Event.parse_obj(payload)
        return EventClass.parse_obj(payload)

    @override
    async def _call_api(self, bot: Bot, api: str, **data: Any) -> Any:
        log("DEBUG", f"Bot {bot.self_id} calling API <y>{api}</y>")
        api_handler: Optional[API] = getattr(bot.__class__, api, None)
        if api_handler is None:
            raise ApiNotAvailable(api)
        return await api_handler(bot, **data)



import re
from functools import partial
from typing_extensions import ParamSpec, Concatenate
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Type,
    Union,
    Generic,
    TypeVar,
    Callable,
    Optional,
    Awaitable,
    overload,
)

from pydantic import Field, BaseModel
from nonebot.utils import logger_wrapper

if TYPE_CHECKING:
    from .bot import Bot

B = TypeVar("B", bound="Bot")
R = TypeVar("R")
P = ParamSpec("P")
log = logger_wrapper("Satori")


def escape(text: str) -> str:
    return text.replace("&", "&amp;").replace('"', "&quot;").replace("<", "&lt;").replace(">", "&gt;")


def unescape(text: str) -> str:
    return text.replace("&quot;", '"').replace("&lt;", "<").replace("&gt;", ">").replace("&amp;", "&")


class Element(BaseModel):
    type: str
    attrs: Dict[str, Any] = Field(default_factory=dict)
    children: List["Element"] = Field(default_factory=list)
    source: Optional[str] = None

    def __str__(self):
        if self.source:
            return self.source
        if self.type == "text":
            return escape(self.attrs["text"])

        def _attr(key: str, value: Any):
            if value is True:
                return key
            if value is False:
                return f"no-{key}"
            if isinstance(value, (int, float)):
                return f"{key}={value}"
            return f'{key}="{escape(str(value))}"'

        attrs = " ".join(_attr(k, v) for k, v in self.attrs.items())
        if not self.children:
            return f"<{self.type} {attrs}/>"
        children = "".join(str(c) for c in self.children)
        return f"<{self.type} {attrs}>{children}</{self.type}>"


tag_pat = re.compile(r"<!--[\s\S]*?-->|<(/?)([^!\s>/]*)([^>]*?)\s*(/?)>")
attr_pat = re.compile(r"([^\s=]+)(?:=\"([^\"]*)\"|='([^']*)')?", re.S)


class Token(BaseModel):
    type: str
    close: str
    empty: str
    attrs: Dict[str, Any]
    source: str


def parse(src: str):
    tokens: List[Union[Token, Element]] = []

    def push_text(text: str):
        if text:
            tokens.append(Element(type="text", attrs={"text": text}))

    def parse_content(source: str):
        push_text(unescape(source))

    while tag_map := tag_pat.search(src):
        parse_content(src[: tag_map.start()])
        src = src[tag_map.end() :]
        if tag_map.group(0).startswith("<!--"):
            continue
        close, tag, attr_str, empty = tag_map.groups()
        tkn = Token(
            type=tag or "template",
            close=close,
            empty=empty,
            attrs={},
            source=tag_map.group(0),
        )
        while attr_map := attr_pat.search(attr_str):
            key, value1, value2 = attr_map.groups()
            value = value1 or value2
            if value:
                tkn.attrs[key] = unescape(value)
            elif key.startswith("no-"):
                tkn.attrs[key] = False
            else:
                tkn.attrs[key] = True
            attr_str = attr_str[attr_map.end() :]
        tokens.append(tkn)

    parse_content(src)

    stack = [Element(type="template")]

    def rollback(i: int):
        while i:
            child = stack.pop(0)
            source = stack[0].children.pop(-1)
            stack[0].children.append(Element(type="text", attrs={"text": source}))
            stack[0].children.extend(child.children)
            i -= 1

    for tkn in tokens:
        if isinstance(tkn, Element):
            stack[0].children.append(tkn)
        elif tkn.close:
            index = 0
            while index < len(stack) and stack[index].type != tkn.type:
                index += 1
            if index == len(stack):
                stack[0].children.append(Element(type="text", attrs={"text": tkn.source}))
            else:
                rollback(index)
                elm = stack.pop(0)
                elm.source = None
        else:
            elm = Element(type=tkn.type, attrs=tkn.attrs)
            stack[0].children.append(elm)
            if not tkn.empty:
                elm.source = tkn.source
                stack.insert(0, elm)

    rollback(len(stack) - 1)
    return stack[0].children


class API(Generic[B, P, R]):
    def __init__(self, func: Callable[Concatenate[B, P], Awaitable[R]]) -> None:
        self.func = func

    def __set_name__(self, owner: Type[B], name: str) -> None:
        self.name = name

    @overload
    def __get__(self, obj: None, objtype: Type[B]) -> "API[B, P, R]":
        ...

    @overload
    def __get__(self, obj: B, objtype: Optional[Type[B]]) -> Callable[P, Awaitable[R]]:
        ...

    def __get__(
        self, obj: Optional[B], objtype: Optional[Type[B]] = None
    ) -> "API[B, P, R] | Callable[P, Awaitable[R]]":
        if obj is None:
            return self

        return partial(obj.call_api, self.name)  # type: ignore

    async def __call__(self, inst: B, *args: P.args, **kwds: P.kwargs) -> R:
        return await self.func(inst, *args, **kwds)



from enum import Enum
from copy import deepcopy
from typing_extensions import override
from typing import TYPE_CHECKING, Any, Dict, Type, TypeVar, Optional

from pydantic import root_validator
from nonebot.utils import escape_tag

from nonebot.adapters import Event as BaseEvent

from .models import Role, User
from .models import Event as SatoriEvent
from .message import Message, RenderMessage
from .models import InnerMessage as SatoriMessage
from .models import ArgvInteraction, ButtonInteraction
from .models import Guild, Login, Channel, ChannelType, InnerMember

E = TypeVar("E", bound="Event")


class EventType(str, Enum):
    FRIEND_REQUEST = "friend-request"
    GUILD_ADDED = "guild-added"
    GUILD_MEMBER_ADDED = "guild-member-added"
    GUILD_MEMBER_REMOVED = "guild-member-removed"
    GUILD_MEMBER_REQUEST = "guild-member-request"
    GUILD_MEMBER_UPDATED = "guild-member-updated"
    GUILD_REMOVED = "guild-removed"
    GUILD_REQUEST = "guild-request"
    GUILD_ROLE_CREATED = "guild-role-created"
    GUILD_ROLE_DELETED = "guild-role-deleted"
    GUILD_ROLE_UPDATED = "guild-role-updated"
    GUILD_UPDATED = "guild-updated"
    LOGIN_ADDED = "login-added"
    LOGIN_REMOVED = "login-removed"
    LOGIN_UPDATED = "login-updated"
    MESSAGE_CREATED = "message-created"
    MESSAGE_DELETED = "message-deleted"
    MESSAGE_UPDATED = "message-updated"
    REACTION_ADDED = "reaction-added"
    REACTION_REMOVED = "reaction-removed"
    INTERNAL = "internal"
    INTERACTION_BUTTON = "interaction/button"
    INTERACTION_COMMAND = "interaction/command"


class Event(BaseEvent, SatoriEvent):
    __type__: EventType

    @override
    def get_type(self) -> str:
        return ""

    @override
    def get_event_name(self) -> str:
        return self.type

    @override
    def get_event_description(self) -> str:
        return escape_tag(str(self.dict()))

    @override
    def get_message(self) -> Message:
        raise ValueError("Event has no message!")

    @override
    def get_user_id(self) -> str:
        if self.user:
            return self.user.id
        raise ValueError("Event has no context!")

    @override
    def get_session_id(self) -> str:
        if self.channel:
            if self.guild:
                return f"{self.guild.id}/{self.channel.id}"
            return self.channel.id
        raise ValueError("Event has no context!")

    @override
    def is_tome(self) -> bool:
        return False


EVENT_CLASSES: Dict[str, Type[Event]] = {}


def register_event_class(event_class: Type[E]) -> Type[E]:
    EVENT_CLASSES[event_class.__type__.value] = event_class
    return event_class


class NoticeEvent(Event):
    @override
    def get_type(self) -> str:
        return "notice"


class FriendEvent(NoticeEvent):
    user: User

    @override
    def get_user_id(self) -> str:
        return self.user.id

    @override
    def get_session_id(self) -> str:
        return self.user.id


@register_event_class
class FriendRequestEvent(FriendEvent):
    __type__ = EventType.FRIEND_REQUEST


class GuildEvent(NoticeEvent):
    guild: Guild

    @override
    def get_session_id(self) -> str:
        return self.guild.id


@register_event_class
class GuildAddedEvent(GuildEvent):
    __type__ = EventType.GUILD_ADDED


@register_event_class
class GuildRemovedEvent(GuildEvent):
    __type__ = EventType.GUILD_REMOVED


@register_event_class
class GuildRequestEvent(GuildEvent):
    __type__ = EventType.GUILD_REQUEST


@register_event_class
class GuildUpdatedEvent(GuildEvent):
    __type__ = EventType.GUILD_UPDATED


class GuildMemberEvent(GuildEvent):
    user: User

    @override
    def get_user_id(self) -> str:
        return self.user.id

    @override
    def get_session_id(self) -> str:
        return f"{self.guild.id}/{self.get_user_id()}"


@register_event_class
class GuildMemberAddedEvent(GuildMemberEvent):
    __type__ = EventType.GUILD_MEMBER_ADDED


@register_event_class
class GuildMemberRemovedEvent(GuildMemberEvent):
    __type__ = EventType.GUILD_MEMBER_REMOVED


@register_event_class
class GuildMemberRequestEvent(GuildMemberEvent):
    __type__ = EventType.GUILD_MEMBER_REQUEST

    member: InnerMember


@register_event_class
class GuildMemberUpdatedEvent(GuildMemberEvent):
    __type__ = EventType.GUILD_MEMBER_UPDATED

    member: InnerMember


class GuildRoleEvent(GuildEvent):
    role: Role

    @override
    def get_session_id(self) -> str:
        return f"{self.guild.id}/{self.role.id}"


@register_event_class
class GuildRoleCreatedEvent(GuildRoleEvent):
    __type__ = EventType.GUILD_ROLE_CREATED


@register_event_class
class GuildRoleDeletedEvent(GuildRoleEvent):
    __type__ = EventType.GUILD_ROLE_DELETED


@register_event_class
class GuildRoleUpdatedEvent(GuildRoleEvent):
    __type__ = EventType.GUILD_ROLE_UPDATED


class LoginEvent(NoticeEvent):
    login: Login


@register_event_class
class LoginAddedEvent(LoginEvent):
    __type__ = EventType.LOGIN_ADDED


@register_event_class
class LoginRemovedEvent(LoginEvent):
    __type__ = EventType.LOGIN_REMOVED


@register_event_class
class LoginUpdatedEvent(LoginEvent):
    __type__ = EventType.LOGIN_UPDATED


class MessageEvent(Event):
    channel: Channel
    user: User
    message: SatoriMessage
    to_me: bool = False
    reply: Optional[RenderMessage] = None

    if TYPE_CHECKING:
        _message: Message
        original_message: Message

    @override
    def get_type(self) -> str:
        return "message"

    @override
    def is_tome(self) -> bool:
        return self.to_me

    @override
    def get_message(self) -> Message:
        return self._message

    @root_validator
    def generate_message(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        values["_message"] = Message.from_satori_element(values["message"].content)
        values["original_message"] = deepcopy(values["_message"])
        return values

    @property
    def msg_id(self) -> str:
        return self.message.id

    def convert(self) -> "MessageEvent":
        raise NotImplementedError


@register_event_class
class MessageCreatedEvent(MessageEvent):
    __type__ = EventType.MESSAGE_CREATED

    def convert(self):
        if self.channel.type == ChannelType.DIRECT:
            return PrivateMessageCreatedEvent.parse_obj(self)
        else:
            return PublicMessageCreatedEvent.parse_obj(self)


@register_event_class
class MessageDeletedEvent(MessageEvent):
    __type__ = EventType.MESSAGE_DELETED

    def convert(self):
        if self.channel.type == ChannelType.DIRECT:
            return PrivateMessageDeletedEvent.parse_obj(self)
        else:
            return PublicMessageDeletedEvent.parse_obj(self)


@register_event_class
class MessageUpdatedEvent(MessageEvent):
    __type__ = EventType.MESSAGE_UPDATED

    def convert(self):
        if self.channel.type == ChannelType.DIRECT:
            return PrivateMessageUpdatedEvent.parse_obj(self)
        else:
            return PublicMessageUpdatedEvent.parse_obj(self)


class PrivateMessageEvent(MessageEvent):
    @override
    def is_tome(self) -> bool:
        return True

    @override
    def get_session_id(self) -> str:
        return self.channel.id

    @override
    def get_user_id(self) -> str:
        return self.user.id


class PublicMessageEvent(MessageEvent):
    member: InnerMember

    @override
    def get_session_id(self) -> str:
        s = f"{self.channel.id}/{self.user.id}"
        if self.guild:
            s = f"{self.guild.id}/{s}"
        return s

    @override
    def get_user_id(self) -> str:
        return self.user.id


class PrivateMessageCreatedEvent(MessageCreatedEvent, PrivateMessageEvent):
    @override
    def get_event_description(self) -> str:
        return escape_tag(
            f"Message {self.msg_id} from "
            f"{self.user.name or ''}({self.channel.id}): {self.get_message()!r}"
        )


class PublicMessageCreatedEvent(MessageCreatedEvent, PublicMessageEvent):
    @override
    def get_event_description(self) -> str:
        return escape_tag(
            f"Message {self.msg_id} from "
            f"{self.user.name or ''}({self.channel.id})"
            f"@[{self.channel.name or ''}:{self.channel.id}]"
            f": {self.get_message()!r}"
        )


class PrivateMessageDeletedEvent(MessageDeletedEvent, PrivateMessageEvent):
    @override
    def get_event_description(self) -> str:
        return escape_tag(f"Message {self.msg_id} from " f"{self.user.name or ''}({self.channel.id}) deleted")


class PublicMessageDeletedEvent(MessageDeletedEvent, PublicMessageEvent):
    @override
    def get_event_description(self) -> str:
        return escape_tag(
            f"Message {self.msg_id} from "
            f"{self.user.name or ''}({self.channel.id})"
            f"@[{self.channel.name or ''}:{self.channel.id}] deleted"
        )


class PrivateMessageUpdatedEvent(MessageUpdatedEvent, PrivateMessageEvent):
    @override
    def get_event_description(self) -> str:
        return escape_tag(
            f"Message {self.msg_id} from "
            f"{self.user.name or ''}({self.channel.id}) updated"
            f": {self.get_message()!r}"
        )


class PublicMessageUpdatedEvent(MessageUpdatedEvent, PublicMessageEvent):
    @override
    def get_event_description(self) -> str:
        return escape_tag(
            f"Message {self.msg_id} from "
            f"{self.user.name or ''}({self.channel.id})"
            f"@[{self.channel.name or ''}:{self.channel.id}] updated"
            f": {self.get_message()!r}"
        )


class ReactionEvent(NoticeEvent):
    channel: Channel
    user: User
    message: SatoriMessage

    if TYPE_CHECKING:
        _message: Message

    @override
    def get_user_id(self) -> str:
        return self.user.id

    @override
    def get_session_id(self) -> str:
        return f"{self.channel.id}/{self.user.id}"

    @root_validator
    def generate_message(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        values["_message"] = Message.from_satori_element(values["message"]["content"])
        return values

    @property
    def msg_id(self) -> str:
        return self.message.id


@register_event_class
class ReactionAddedEvent(ReactionEvent):
    __type__ = EventType.REACTION_ADDED

    @override
    def get_event_description(self) -> str:
        return escape_tag(f"Reaction added to {self.msg_id} by {self.user.name}({self.channel.id})")


@register_event_class
class ReactionRemovedEvent(ReactionEvent):
    __type__ = EventType.REACTION_REMOVED

    @override
    def get_event_description(self) -> str:
        return escape_tag(f"Reaction removed from {self.msg_id}")


@register_event_class
class InternalEvent(Event):
    __type__ = EventType.INTERNAL

    @override
    def get_event_name(self) -> str:
        return getattr(self, "_type", "internal")


class InteractionEvent(NoticeEvent):
    def convert(self) -> "InteractionEvent":
        raise NotImplementedError

    @override
    def is_tome(self) -> bool:
        return True


@register_event_class
class InteractionButtonEvent(InteractionEvent):
    __type__ = EventType.INTERACTION_BUTTON

    button: ButtonInteraction

    @override
    def get_event_description(self) -> str:
        return escape_tag(f"Button interacted with button#{self.button.id}")

    def convert(self):
        if self.channel and self.user and self.channel.type != ChannelType.DIRECT:
            return PublicInteractionButtonEvent.parse_obj(self)
        if self.user:
            return PrivateInteractionButtonEvent.parse_obj(self)
        return self


class PrivateInteractionButtonEvent(InteractionButtonEvent):
    user: User

    @override
    def get_session_id(self) -> str:
        return self.channel.id if self.channel else self.user.id

    @override
    def get_user_id(self) -> str:
        return self.user.id


class PublicInteractionButtonEvent(InteractionButtonEvent):
    user: User
    channel: Channel

    @override
    def get_session_id(self) -> str:
        s = f"{self.channel.id}/{self.user.id}"
        if self.guild:
            s = f"{self.guild.id}/{s}"
        return s

    @override
    def get_user_id(self) -> str:
        return self.user.id


@register_event_class
class InteractionCommandEvent(InteractionEvent):
    __type__ = EventType.INTERACTION_COMMAND

    if TYPE_CHECKING:
        _message: Message
        original_message: Message

    @override
    def get_type(self) -> str:
        return "message"

    @override
    def get_message(self) -> Message:
        return self._message

    def convert(self):
        if self.argv:
            return InteractionCommandArgvEvent.convert(self)  # type: ignore
        return InteractionCommandMessageEvent.convert(self)  # type: ignore


class InteractionCommandArgvEvent(InteractionCommandEvent):
    argv: ArgvInteraction

    @override
    def get_event_description(self) -> str:
        return escape_tag(f"Command interacted with {self.argv}")

    @root_validator
    def generate_message(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        argv: ArgvInteraction = values["argv"]
        cmd = argv.name
        if argv.arguments:
            cmd += " ".join(argv.arguments)
        values["_message"] = Message(cmd)
        values["original_message"] = deepcopy(values["_message"])
        return values

    def convert(self):
        if self.channel and self.user and self.channel.type != ChannelType.DIRECT:
            return PublicInteractionCommandArgvEvent.parse_obj(self)
        if self.user:
            return PrivateInteractionCommandArgvEvent.parse_obj(self)
        return self


class PrivateInteractionCommandArgvEvent(InteractionCommandArgvEvent):
    user: User

    @override
    def get_session_id(self) -> str:
        return self.channel.id if self.channel else self.user.id

    @override
    def get_user_id(self) -> str:
        return self.user.id


class PublicInteractionCommandArgvEvent(InteractionCommandArgvEvent):
    user: User
    channel: Channel

    @override
    def get_session_id(self) -> str:
        s = f"{self.channel.id}/{self.user.id}"
        if self.guild:
            s = f"{self.guild.id}/{s}"
        return s

    @override
    def get_user_id(self) -> str:
        return self.user.id


class InteractionCommandMessageEvent(InteractionCommandEvent):
    message: SatoriMessage
    to_me: bool = False
    reply: Optional[RenderMessage] = None

    @root_validator
    def generate_message(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        values["_message"] = Message.from_satori_element(values["message"].content)
        values["original_message"] = deepcopy(values["_message"])
        return values

    @override
    def get_event_description(self) -> str:
        return escape_tag(f"Command interacted with {self.get_message()}")

    def convert(self):
        if self.channel and self.user and self.channel.type != ChannelType.DIRECT:
            return PublicInteractionCommandMessageEvent.parse_obj(self)
        if self.user:
            return PrivateInteractionCommandMessageEvent.parse_obj(self)
        return self


class PrivateInteractionCommandMessageEvent(InteractionCommandMessageEvent):
    user: User

    @override
    def get_session_id(self) -> str:
        return self.channel.id if self.channel else self.user.id

    @override
    def get_user_id(self) -> str:
        return self.user.id


class PublicInteractionCommandMessageEvent(InteractionCommandMessageEvent):
    user: User
    channel: Channel

    @override
    def get_session_id(self) -> str:
        s = f"{self.channel.id}/{self.user.id}"
        if self.guild:
            s = f"{self.guild.id}/{s}"
        return s

    @override
    def get_user_id(self) -> str:
        return self.user.id
